﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Runtime.Serialization;
using Uploader;
using System.Threading;
using System.Reflection;
using Common.Ini;

namespace Logger
{
    public class Program
    {
        const string LOG_FOLDER = @"PacketLog"; // default to this folder inside current folder
        static string fullPathToLogFolder;

        public static string FullPathToLogFolder()
        {
            var currentFolder = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            var defaultFolder = Path.Combine(currentFolder, LOG_FOLDER);
            var iniFile = new IniFile();
            var logFolder = iniFile.IniReadValue("FOLDERS", "LogFolder");
            if (string.IsNullOrWhiteSpace(logFolder))
            {
                return defaultFolder;
            }
            return logFolder;
        }

        static void Main(string[] args)
        {
            fullPathToLogFolder = FullPathToLogFolder();
            Directory.CreateDirectory(fullPathToLogFolder); // creates if it doesn't exist

            Console.WriteLine("Running...");
            Console.WriteLine("Press any key to exit.");

            using (var serialPortController = new SerialPortController("COM1"))
            {
                var fcPacketGetter = new PacketGetter(serialPortController, 0xFC);
                var fdPacketGetter = new PacketGetter(serialPortController, 0xFD);

                var fcErrorCount = 0;
                var fdErrorCount = 0;
                var toggle = false;
                while (!Console.KeyAvailable)
                {
                    if (toggle)
                    {
                        logPacket("FC", fcPacketGetter, ref fcErrorCount);
                        if (fcErrorCount > 0)
                        {
                            serialPortController.ClosePortIfOpen();
                            Thread.Sleep(1000);
                        }
                    }
                    else
                    {
                        logPacket("FD", fdPacketGetter, ref fdErrorCount);
                        if (fdErrorCount > 0)
                        {
                            serialPortController.ClosePortIfOpen();
                            Thread.Sleep(1000);
                        }
                    }
                    Thread.Sleep(3500);
                    toggle = !toggle;
                }
            }
            Console.WriteLine("Exited.");
        }

        private static void logPacket(string commandByte, PacketGetter packetGetter, ref int errorCount)
        {
            Exception exception;
            if (tryLogPacket(packetGetter, out exception))
            {
                errorCount = 0;
                Console.WriteLine("Logged " + commandByte);
            }
            else
            {
                errorCount++;
                logError(errorCount, commandByte, exception);
            }
        }

        private static bool tryLogPacket(PacketGetter packetGetter, out Exception exception)
        {
            PolarizedPacket polarizedPacket;
            Exception ex;
            var result = packetGetter.TryReadPolarizedPacket(out polarizedPacket, out ex);
            if (result)
            {
                if (polarizedPacket.PacketValid)
                {
                    var filename = polarizedPacket.GenerateFilename();
                    var fullPathToFile = Path.Combine(fullPathToLogFolder, filename);
                    polarizedPacket.SerializeObject(fullPathToFile);
                    exception = null;
                }
                else
                {
                    result = false;
                    exception = new Exception(polarizedPacket.ValidationError);
                }
            }
            else
            {
                exception = ex;
            }
            return result;
        }

        private static void logError(int errorCount, string commandByte, Exception exception)
        {
            Console.WriteLine("Error logging from " + commandByte + ":");
            Console.WriteLine(exception.Message);
        }
    }
}
