﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Common.Ini;
using PacketProcessor.Email;
using System.Threading;

namespace PacketProcessor.Alarms
{
    class AlarmManager
    {
        private readonly List<IAlarm> alarms = new List<IAlarm>();
        private readonly Dictionary<IAlarm, bool> alarmActive = new Dictionary<IAlarm, bool>();
        private readonly GmailEmailer emailer;

        public AlarmManager()
        {
            var iniFile = new IniFile();
            this.emailer = new GmailEmailer(
                iniFile.IniReadValue("ALARMS", "AlarmEmail"),
                iniFile.IniReadValue("ALARMS", "AlarmEmailPassword"));
            this.loadAlarms(iniFile);
            this.printAlarms();
        }

        private void loadAlarms(IniFile iniFile)
        {
            var alarmCountString = iniFile.IniReadValue("ALARMS", "AlarmCount");
            int alarmCount;
            if (int.TryParse(alarmCountString, out alarmCount) && alarmCount > 0)
            {
                for (var alarmNumber = 1; alarmNumber <= alarmCount; alarmNumber++)
                {
                    var index = string.Format("{0:000}", alarmNumber);
                    var ok = true;
                    var description = iniFile.IniReadValue("ALARMS", "AlarmDescription" + index) ?? string.Empty;
                    var channelString = iniFile.IniReadValue("ALARMS", "AlarmChannel" + index);
                    byte channel;
                    if (!byte.TryParse(channelString, out channel))
                    {
                        ok = false;
                    }
                    var thresholdWattsString = iniFile.IniReadValue("ALARMS", "AlarmWatts" + index);
                    int thresholdWatts;
                    if (!int.TryParse(thresholdWattsString, out thresholdWatts))
                    {
                        ok = false;
                    }
                    if (ok)
                    {
                        var alarm = new Alarm
                        {
                            AlarmNumber = alarmNumber,
                            Description = description,
                            Channel = channel,
                            ThresholdWatts = thresholdWatts,
                        };
                        this.alarms.Add(alarm);
                        this.alarmActive.Add(alarm, false);
                    }
                }
            }
        }

        private void printAlarms()
        {
            Console.WriteLine("Loaded alarm definitions:");
            foreach (var alarm in this.alarms)
            {
                Console.WriteLine(alarm);
            }
        }

        public void ProcessAlarms(TimeSlice timeSlice)
        {
            foreach (var alarm in this.alarms)
            {
                this.processAlarm(alarm, timeSlice);
            }
        }

        private void processAlarm(IAlarm alarm, TimeSlice timeSlice)
        {
            var active = alarm.IsActive(timeSlice);
            var previousState = this.alarmActive[alarm];
            if (active && !previousState)
            {
                // alarm became active
                var alarmText = "Alarm triggered: " + alarm.ToString();
                Console.WriteLine(alarmText);
                ThreadPool.QueueUserWorkItem(o =>
                    this.emailer.TryEmail(alarmText, string.Empty));
                this.alarmActive[alarm] = true;
            }
            else if(previousState && !active)
            {
                // alarm no longer active
                this.alarmActive[alarm] = false;
            }
        }
    }
}
