﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Uploader;

namespace PacketProcessor.ComparePackets
{
    class PacketComparer : IPacketComparer
    {
        public IPacketComparison Compare(PolarizedPacket oldPacket, PolarizedPacket newPacket)
        {
            return new PacketComparison
            {
                Received = newPacket.Received,
                SecondsElapsed = secondsElapsed(oldPacket, newPacket),
                Channel1Diff = channel1Or2Diff(oldPacket.Channel1PolarizedWattSeconds, newPacket.Channel1PolarizedWattSeconds),
                Channel2Diff = channel1Or2Diff(oldPacket.Channel2PolarizedWattSeconds, newPacket.Channel2PolarizedWattSeconds),
                Aux1Diff = auxChannelDiff(oldPacket.ChannelAux1WattSeconds, newPacket.ChannelAux1WattSeconds),
                Aux2Diff = auxChannelDiff(oldPacket.ChannelAux2WattSeconds, newPacket.ChannelAux2WattSeconds),
                Aux3Diff = auxChannelDiff(oldPacket.ChannelAux3WattSeconds, newPacket.ChannelAux3WattSeconds),
                Aux4Diff = auxChannelDiff(oldPacket.ChannelAux4WattSeconds, newPacket.ChannelAux4WattSeconds),
                Aux5Diff = auxChannelDiff(oldPacket.ChannelAux5WattSeconds, newPacket.ChannelAux5WattSeconds),
            };
        }

        private static uint secondsElapsed(PolarizedPacket oldPacket, PolarizedPacket newPacket)
        {
            const int ROLLOVER = 256 * 256 * 256;
            var result = newPacket.SecondsCounter - oldPacket.SecondsCounter;
            if (result < (-ROLLOVER / 2))
            {
                result += ROLLOVER;
            }
            if (result < 0) return 0;
            return (uint)result;
        }

        private static uint channel1Or2Diff(long oldWattSeconds, long newWattSeconds)
        {
            const long CH_12_ROLLOVER = 256L * 256L * 256L * 256L * 256L;

            var result = newWattSeconds - oldWattSeconds;
            if (result < (-CH_12_ROLLOVER / 2))
            {
                result += CH_12_ROLLOVER;
            }
            if (result < 0) return 0;
            return (uint)result;
        }

        private static uint auxChannelDiff(long oldWattSeconds, long newWattSeconds)
        {
            const long CH_AUX_ROLLOVER = 256L * 256L * 256L * 256L;

            var result = newWattSeconds - oldWattSeconds;
            if (result < (-CH_AUX_ROLLOVER / 2))
            {
                result += CH_AUX_ROLLOVER;
            }
            if (result < 0) return 0;
            return (uint)result;
        }

        public uint[] WattSecondsInTimeSlice(IPacketComparison packetComparison, DateTime startOfSlice, DateTime endOfSlice)
        {
            var packetSeconds = packetComparison.SecondsElapsed;
            var secondsInSlice = calculateSecondsInTimeSlice(packetComparison, startOfSlice, endOfSlice);
            var result = new uint[7];
            if (packetSeconds > 0)
            {
                result[0] = packetComparison.Channel1Diff * secondsInSlice / packetSeconds;
                result[1] = packetComparison.Channel2Diff * secondsInSlice / packetSeconds;
                result[2] = packetComparison.Aux1Diff * secondsInSlice / packetSeconds;
                result[3] = packetComparison.Aux2Diff * secondsInSlice / packetSeconds;
                result[4] = packetComparison.Aux3Diff * secondsInSlice / packetSeconds;
                result[5] = packetComparison.Aux4Diff * secondsInSlice / packetSeconds;
                result[6] = packetComparison.Aux5Diff * secondsInSlice / packetSeconds;
            }
            return result;
        }

        private static uint calculateSecondsInTimeSlice(IPacketComparison packetComparison, DateTime startOfSlice, DateTime endOfSlice)
        {
            var received = packetComparison.Received;
            if (received <= startOfSlice)
            {
                return 0;
            }
            var secondsSinceStart = Convert.ToUInt32(received.Subtract(startOfSlice).TotalSeconds);
            if (received <= endOfSlice)
            {
                return Math.Min(packetComparison.SecondsElapsed, secondsSinceStart);
            }
            var secondsSinceEnd = Convert.ToUInt32(received.Subtract(endOfSlice).TotalSeconds);
            if (packetComparison.SecondsElapsed < secondsSinceEnd)
            {
                return 0;
            }
            var result = packetComparison.SecondsElapsed - secondsSinceEnd;
            return Math.Min(result, TimeSlice.SECONDS_PER_SLICE);
        }
    }
}
