﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Collections.ObjectModel;
using Uploader;
using PacketProcessor.ComparePackets;
using Common.Ini;
using PacketProcessor.Alarms;

namespace PacketProcessor
{
    public class Program
    {
        const int DEVICE_1_SERIAL = 500133;
        const int DEVICE_2_SERIAL = 500134;
        const ushort SLICES_PER_DAY = 8640; // 24 * 60 * 6 (one slice is 10 seconds)
        public const uint FULL_TIMESLICE_FILE_LENGTH = TimeSlice.BYTES_PER_SLICE * SLICES_PER_DAY;

        const string DATA_FOLDER = @"Data"; // default to this folder inside current folder
        static string fullPathToLogFolder;
        static string fullPathToDataFolder;
        static AlarmManager alarmManager;

        public static string FullPathToDataFolder()
        {
            var currentFolder = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            var defaultFolder = Path.Combine(currentFolder, DATA_FOLDER);
            var iniFile = new IniFile();
            var dataFolder = iniFile.IniReadValue("FOLDERS", "DataFolder");
            if (string.IsNullOrWhiteSpace(dataFolder))
            {
                return defaultFolder;
            }
            return dataFolder;
        }

        static void Main(string[] args)
        {
            fullPathToLogFolder = Logger.Program.FullPathToLogFolder();
            fullPathToDataFolder = FullPathToDataFolder();
            Directory.CreateDirectory(fullPathToLogFolder); // creates if it doesn't exist
            Directory.CreateDirectory(fullPathToDataFolder); // creates if it doesn't exist

            alarmManager = new AlarmManager();

            Console.WriteLine("Running...");
            Console.WriteLine("Press any key to exit.");

            var secondCount = 0;
            while (!Console.KeyAvailable)
            {
                Thread.Sleep(100);
                secondCount++;
                if (secondCount >= 1)
                {
                    secondCount = 0;
                    processPacketQueue();
                }
            }
            
        }

        private static void processPacketQueue()
        {
            var packetQueue = new PacketQueue(fullPathToLogFolder);
            var packets = packetQueue.Packets;
            if (packets.Count > 0)
            {
                processPackets(packets);
            }
        }

        private static void processPackets(ReadOnlyCollection<PolarizedPacket> packets)
        {
            var filename = SliceFileFinder.FindLatestSliceFile(fullPathToDataFolder);
            var fileInfo = new FileInfo(filename);
            var sliceNumber = fileInfo.Length / TimeSlice.BYTES_PER_SLICE;
            var keepGoing = true;
            while (keepGoing && sliceNumber < SLICES_PER_DAY)
            {
                keepGoing = createSlice(filename, (ushort)sliceNumber, packets);
                sliceNumber++;
            }
        }

        // Returns true if slice created
        private static bool createSlice(
            string filename, 
            ushort sliceNumber,
            ReadOnlyCollection<PolarizedPacket> packets)
        {
            var result = false;
            IPacketComparer packetComparer = new PacketComparer();
            var date = DateTime.Parse(Path.GetFileNameWithoutExtension(filename));
            var startOfSlice = date.AddSeconds(sliceNumber * TimeSlice.SECONDS_PER_SLICE);
            var endOfSlice = startOfSlice.AddSeconds(TimeSlice.SECONDS_PER_SLICE);

            // need at least one packet from each device where received time > endOfSlice
            var device1Packets = packets
                .Where(x => x.SerialNumber == DEVICE_1_SERIAL)
                .ToList();
            var device2Packets = packets
                .Where(x => x.SerialNumber == DEVICE_2_SERIAL)
                .ToList();
            if (device1Packets.Any(x => x.Received > endOfSlice)
                && device2Packets.Any(x => x.Received > endOfSlice))
            {
                var voltage = device1Packets
                    .Where(x => x.Received >= startOfSlice)
                    .First()
                    .Voltage;
                var wattSeconds = new uint[TimeSlice.CHANNELS];

                {
                    var previousDevice1Packet = device1Packets[0];
                    for (var i = 1; i < device1Packets.Count; i++)
                    {
                        var device1Packet = device1Packets[i];
                        var comparison = packetComparer.Compare(previousDevice1Packet, device1Packet);
                        var seconds = comparison.SecondsElapsed;
                        var channelDiffs = packetComparer.WattSecondsInTimeSlice(comparison, startOfSlice, endOfSlice);
                        for (var c = 0; c < 7; c++)
                        {
                            wattSeconds[c] += channelDiffs[c];
                        }
                        if (device1Packet.Received > endOfSlice)
                        {
                            break;
                        }
                        previousDevice1Packet = device1Packet;
                    }
                }

                {
                    var previousDevice2Packet = device2Packets[0];
                    for (var i = 1; i < device2Packets.Count; i++)
                    {
                        var device2Packet = device2Packets[i];
                        var comparison = packetComparer.Compare(previousDevice2Packet, device2Packet);
                        var seconds = comparison.SecondsElapsed;
                        var channelDiffs = packetComparer.WattSecondsInTimeSlice(comparison, startOfSlice, endOfSlice);
                        for (var c = 0; c < 7; c++)
                        {
                            wattSeconds[c+7] += channelDiffs[c];
                        }
                        if (device2Packet.Received > endOfSlice)
                        {
                            break;
                        }
                        previousDevice2Packet = device2Packet;
                    }
                }

                var timeSlice = new TimeSlice(voltage, wattSeconds);
                Console.WriteLine("Creating slice for " + date.ToString("yyyy-MM-dd") + " slice number " + sliceNumber + " to " + endOfSlice.ToString("HH:mm:ss"));
                AppendAllBytes(filename, timeSlice.GetBytes());
                result = true;
                alarmManager.ProcessAlarms(timeSlice);
                if (sliceNumber == 0)
                {
                    var device1Packet = device1Packets.First(x => x.Received >= startOfSlice);
                    var device2Packet = device2Packets.First(x => x.Received >= startOfSlice);
                    writeDailyCounters(device1Packet, device2Packet);
                }
            }
            deleteOldPackets(device1Packets, startOfSlice);
            deleteOldPackets(device2Packets, startOfSlice);
            return result;
        }

        private static void deleteOldPackets(List<PolarizedPacket> packets, DateTime startOfSlice)
        {
            var packetsBeforeEndOfSlice = packets
                .Where(x => x.Received < startOfSlice)
                .ToList();
            if (packetsBeforeEndOfSlice.Count > 1)
            {
                var toBeDeleted = packetsBeforeEndOfSlice
                    .Take(packetsBeforeEndOfSlice.Count - 1)
                    .ToList();
                foreach (var packet in toBeDeleted)
                {
                    deletePacket(packet);
                }
            }
        }

        private static void deletePacket(PolarizedPacket packet)
        {
            var filename = packet.GenerateFilename();
            var fullPath = Path.Combine(fullPathToLogFolder, filename);
            if (File.Exists(fullPath))
            {
                File.Delete(fullPath);
            }
        }

        public static void AppendAllBytes(string path, byte[] bytes)
        {
            var success = false;
            var tryCount = 0;
            while (!success && tryCount < 3)
            {
                try
                {
                    tryCount++;
                    using (var stream = new FileStream(path, FileMode.Append))
                    {
                        stream.Write(bytes, 0, bytes.Length);
                        success = true;
                    }
                }
                catch (IOException)
                {
                    Thread.Sleep(1000);
                }
            }
        }

        private static void writeDailyCounters(PolarizedPacket device1packet, PolarizedPacket device2packet)
        {
            var filename = Path.Combine(fullPathToDataFolder, "DailyCounters.csv");
            if (!File.Exists(filename))
            {
                SliceFileFinder.MakeZeroByteFile(filename);
                File.AppendAllText(filename, "Device 1 Rcvd,Ch1,Ch2,Aux1,Aux2,Aux3,Aux4,Aux5,Device 2 Rcvd,Ch1,Ch2,Aux1,Aux2,Aux3,Aux4,Aux5");
                File.AppendAllText(filename, Environment.NewLine);
            }
            writePacketDataToCsvFile(device1packet, filename);
            File.AppendAllText(filename, ",");
            writePacketDataToCsvFile(device2packet, filename); 
            File.AppendAllText(filename, Environment.NewLine);
        }

        private static void writePacketDataToCsvFile(PolarizedPacket packet, string filename)
        {
            var text = string.Format("{0},{1},{2},{3},{4},{5},{6},{7}",
                packet.Received.ToString("yyyy-MM-dd HH:mm:ss"),
                packet.Channel1PolarizedWattSeconds,
                packet.Channel2PolarizedWattSeconds,
                packet.ChannelAux1WattSeconds,
                packet.ChannelAux2WattSeconds,
                packet.ChannelAux3WattSeconds,
                packet.ChannelAux4WattSeconds,
                packet.ChannelAux5WattSeconds);
            File.AppendAllText(filename, text);
        }
    }
}
