﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace PacketProcessor
{
    public class TimeSlice
    {
        public const ushort SECONDS_PER_SLICE = 10;
        public const int BYTES_PER_SLICE = 36;
        public const int CHANNELS = 14;
        const uint MAX_WATT_SECONDS_PER_SLICE = 1048575;
        const ulong MULTIPLIER = MAX_WATT_SECONDS_PER_SLICE + 1; // shifts left 20 bits

        private readonly byte[] slice = new byte[BYTES_PER_SLICE]; // byte 0 is voltage, then 20 bits (2.5 bytes) per channel
        private readonly uint[] ws = new uint[CHANNELS];

        public TimeSlice(
            decimal voltage,
            uint[] wattSeconds)
        {
            if (voltage > 255) throw new ArgumentOutOfRangeException("voltage", "Voltage too high (exceeds 255): " + voltage);
            if (voltage < 0) throw new ArgumentOutOfRangeException("voltage", "Voltage too low (lower than 0): " + voltage);
            if (wattSeconds == null) throw new ArgumentNullException("wattSeconds");
            if (wattSeconds.Length != CHANNELS) throw new ArgumentOutOfRangeException("wattSeconds", "Should have " + CHANNELS + " elements, has: " + wattSeconds.Length);
            if (wattSeconds.Any(x => x > MAX_WATT_SECONDS_PER_SLICE)) throw new ArgumentOutOfRangeException("wattSeconds", "Individual channel > max");
            if (!BitConverter.IsLittleEndian) throw new Exception("Expected little-endian architecture.");
            Array.Copy(wattSeconds, this.ws, CHANNELS);
            this.compact(voltage);
        }

        public TimeSlice(byte[] fromBytes)
        {
            if (fromBytes == null) throw new ArgumentNullException("fromBytes");
            if (fromBytes.Length != BYTES_PER_SLICE) throw new ArgumentNullException("fromBytes", "Length should be " + BYTES_PER_SLICE + " bytes, but is " + fromBytes.Length);
            Array.Copy(fromBytes, this.slice, BYTES_PER_SLICE);
            this.expand();
        }

        private void compact(decimal voltage)
        {
            this.slice[0] = Convert.ToByte(voltage);
            for (var i = 0; i < CHANNELS; i+=2)
            {
                // process 2 channels at a time into 5 bytes (2.5 bytes each = 20 bits)
                ulong twoChannels = this.ws[i+1];
                twoChannels += this.ws[i] * MULTIPLIER;

                var destination = i/2 * 5 + 1;
                var bytes = BitConverter.GetBytes(twoChannels);
                // assume little-endian, so index 0 of array is LSB
                Array.Copy(bytes, 0, this.slice, destination, 5);
            }
        }

        private void expand()
        {
            for (var i = 0; i < CHANNELS; i += 2)
            {
                var source = i/2 * 5 + 1;
                var bytes = new byte[8];
                Array.Copy(this.slice, source, bytes, 0, 5);
                var twoChannels = BitConverter.ToUInt64(bytes, 0);
                this.ws[i] = (uint)(twoChannels >> 20);
                this.ws[i+1] = (uint)(twoChannels & 0xFFFFF);
            }
        }

        public byte[] GetBytes()
        {
            return this.slice;
        }

        public byte Voltage { get { return this.slice[0]; } }

        public uint WattSeconds(int channel)
        {
            if (channel < 0) throw new ArgumentOutOfRangeException("channel", "Channel must be >= 0");
            if (channel >= CHANNELS) throw new ArgumentOutOfRangeException("channel", "Channel must be < " + CHANNELS);
            return this.ws[channel];
        }
    }
}
