﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Uploader
{
    [Serializable]
    public class PolarizedPacket
    {
        public const int PACKET_LENGTH = 65;
        private readonly DateTime received;
        private readonly byte[] packetData = new byte[PACKET_LENGTH];

        public PolarizedPacket(DateTime received, byte[] packetData)
        {
            if (packetData == null) throw new ArgumentNullException("packetData");
            this.received = received;
            string validationError;
            if (packetValid(packetData, out validationError))
            {
                Array.Copy(packetData, this.packetData, PACKET_LENGTH);
                this.PacketValid = true;
                this.ValidationError = string.Empty;
            }
            else
            {
                this.PacketValid = false;
                this.ValidationError = validationError;
            }
        }

        public DateTime Received { get { return this.received; } }

        public decimal Voltage
        {
            get
            {
                var tenthsOfVolt = this.packetData[3] * 256 + this.packetData[4];
                return Convert.ToDecimal(tenthsOfVolt) / 10M;
            }
        }

        public long Channel1AbsoluteWattSeconds
        {
            get { return get5ByteWattSeconds(5); }
        }

        public long Channel2AbsoluteWattSeconds
        {
            get { return get5ByteWattSeconds(10); }
        }

        public long Channel1PolarizedWattSeconds
        {
            get { return get5ByteWattSeconds(15); }
        }

        public long Channel2PolarizedWattSeconds
        {
            get { return get5ByteWattSeconds(20); }
        }

        private long get5ByteWattSeconds(int initialAddress)
        {
            var accumulator = 0L;
            accumulator += this.packetData[initialAddress + 0];
            accumulator += this.packetData[initialAddress + 1] * 256L;
            accumulator += this.packetData[initialAddress + 2] * 256L * 256L;
            accumulator += this.packetData[initialAddress + 3] * 256L * 256L * 256L;
            accumulator += this.packetData[initialAddress + 4] * 256L * 256L * 256L * 256L;
            return accumulator;
        }

        public int SerialNumber
        {
            get
            {
                var serialNumber = this.packetData[32] * 100000
                    + this.packetData[29] 
                    + this.packetData[30] * 256;
                return serialNumber;
            }
        }

        /// <summary>
        /// This is a 3-bit counter that increments every time the counters in the
        /// device are reset.  It counts to 7 and rolls over to 0.
        /// "Resetting the ECM-1220 causes all KWh, elapsed time, and data-logger memory to be zeroed." (sic)
        /// </summary>
        public byte ResetCounter
        {
            get
            {
                return (byte)(this.packetData[31] & 0x03);
            }
        }

        /// <summary>
        /// Max value of this counter is 16777215 (i.e. 256*256*256-1)
        /// and then it rolls back over to 0.  Approx 194 days.
        /// </summary>
        public int SecondsCounter
        {
            get
            {
                var accumulator = 0;
                accumulator += this.packetData[37 + 0];
                accumulator += this.packetData[37 + 1] * 256;
                accumulator += this.packetData[37 + 2] * 256 * 256;
                return accumulator;
            }
        }

        public decimal Channel1Current
        {
            get
            {
                var hundredthsOfAmp = this.packetData[33] + this.packetData[34] * 256;
                return Convert.ToDecimal(hundredthsOfAmp) / 100M;
            }
        }

        public decimal Channel2Current
        {
            get
            {
                var hundredthsOfAmp = this.packetData[35] + this.packetData[36] * 256;
                return Convert.ToDecimal(hundredthsOfAmp) / 100M;
            }
        }

        public long ChannelAux1WattSeconds
        {
            get { return get4ByteWattSeconds(40); }
        }

        public long ChannelAux2WattSeconds
        {
            get { return get4ByteWattSeconds(44); }
        }

        public long ChannelAux3WattSeconds
        {
            get { return get4ByteWattSeconds(48); }
        }

        public long ChannelAux4WattSeconds
        {
            get { return get4ByteWattSeconds(52); }
        }

        public long ChannelAux5WattSeconds
        {
            get { return get4ByteWattSeconds(56); }
        }

        private long get4ByteWattSeconds(int initialAddress)
        {
            var accumulator = 0L;
            accumulator += this.packetData[initialAddress + 0];
            accumulator += this.packetData[initialAddress + 1] * 256;
            accumulator += this.packetData[initialAddress + 2] * 256 * 256;
            accumulator += this.packetData[initialAddress + 3] * 256 * 256 * 256;
            return accumulator;
        }

        public bool PacketValid { get; private set; }
        public string ValidationError { get; private set; }

        private static bool packetValid(byte[] packetData, out string validationError)
        {
            if (packetData.Length < PACKET_LENGTH)
            {
                validationError = "Packet length too short.";
                return false;
            }
            if (packetData.Length > PACKET_LENGTH)
            {
                validationError = "Packet length too long.";
                return false;
            }
            if (packetData[0] != 0xFE)
            {
                validationError = "First byte not equal to 0xFE.";
                return false;
            }
            if (packetData[1] != 0xFF)
            {
                validationError = "Second byte not equal to 0xFF.";
                return false;
            }
            if (packetData[2] != 0x03)
            {
                validationError = "Third byte not equal to 0x03.";
                return false;
            }
            if (packetData[62] != 0xFF)
            {
                validationError = "63rd byte not equal to 0xFF.";
                return false;
            }
            if (packetData[63] != 0xFE)
            {
                validationError = "64th byte not equal to 0xFE.";
                return false;
            }
            var sumOfFirst64Bytes = 0;
            for (var i = 0; i <= 63; i++)
            {
                sumOfFirst64Bytes += packetData[i];
            }
            var byte65ShouldBe = (byte)(sumOfFirst64Bytes % 256);
            if (packetData[64] != byte65ShouldBe)
            {
                validationError = "Checksum incorrect.  Should be " + byte65ShouldBe + ", but is " + packetData[64];
                return false;
            }
            validationError = string.Empty;
            return true;
        }

        public byte[] GetBytes()
        {
            if (!this.PacketValid) throw new Exception("Packet data is not valid.");
            return this.packetData;
        }

        public string GenerateFilename()
        {
            return this.Received.ToString("yyyy-MM-dd HHmmss ") + this.SerialNumber + ".xml";
        }

        public override string ToString()
        {
            if (this.PacketValid)
            {
                var result = new StringBuilder();
                result.AppendLine("Serial #: {0}", this.SerialNumber);
                result.AppendLine("Received: {0}", this.Received.ToString("yyyy-MM-dd HH:mm:ss"));
                result.AppendLine("Voltage (V): {0:#.0}", this.Voltage);
                result.AppendLine("Ch. 1 Current (A): {0:#.00}", this.Channel1Current);
                result.AppendLine("Ch. 2 Current (A): {0:#.00}", this.Channel2Current);
                result.AppendLine("Seconds Counter: {0}", this.SecondsCounter);
                result.AppendLine("Reset Counter: {0}", this.ResetCounter);
                result.AppendLine("Ch. 1 Absolute WS: {0}", this.Channel1AbsoluteWattSeconds);
                result.AppendLine("Ch. 2 Absolute WS: {0}", this.Channel2AbsoluteWattSeconds);
                result.AppendLine("Ch. 1 Polarized WS: {0}", this.Channel1PolarizedWattSeconds);
                result.AppendLine("Ch. 2 Polarized WS: {0}", this.Channel2PolarizedWattSeconds);
                result.AppendLine("Ch. Aux 1 WS: {0}", this.ChannelAux1WattSeconds);
                result.AppendLine("Ch. Aux 2 WS: {0}", this.ChannelAux2WattSeconds);
                result.AppendLine("Ch. Aux 3 WS: {0}", this.ChannelAux3WattSeconds);
                result.AppendLine("Ch. Aux 4 WS: {0}", this.ChannelAux4WattSeconds);
                result.AppendLine("Ch. Aux 5 WS: {0}", this.ChannelAux5WattSeconds);
                return result.ToString();
            }
            else
            {
                return this.ValidationError;
            }
        }
    }
}
