﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Threading;

namespace Uploader
{
    public class SerialPortController : ISerialPortController
    {
        const int MAX_LENGTH = 256;
        const int RECEIVE_TIMEOUT_MS = 2000;

        private readonly string comPortName;
        private SerialPort port;
        private readonly IncomingBuffer buffer = new IncomingBuffer();

        public SerialPortController(string comPortName)
        {
            if (comPortName == null) throw new ArgumentNullException("comPortName");
            this.comPortName = comPortName;
        }

        public byte[] SendReceive(byte[] send, int receiveLength)
        {
            byte[] result = new byte[0];
            serialPortTransaction(() =>
            {
                this.buffer.Clear();
                this.port.Write(send, 0, send.Length);
                result = waitForMessageByLength(receiveLength);
            });
            return result;
        }

        private byte[] waitForMessageByLength(int receiveLength)
        {
            receiveToBuffer();
            int msWaited = 0;
            while (this.buffer.Length < receiveLength)
            {
                if (msWaited >= RECEIVE_TIMEOUT_MS) throw new TimeoutException("Timeout waiting for receipt of data.");
                Thread.Sleep(10); msWaited += 10;
                receiveToBuffer();
            }
            var result = this.buffer.GetBytes();
            this.buffer.Clear();
            return result;
        }

        /// <summary>
        /// Copies bytes from the serial port into the buffer.
        /// Should be called inside a serial port transaction.
        /// </summary>
        private void receiveToBuffer()
        {
            byte[] incoming = new byte[MAX_LENGTH];
            if (this.port.BytesToRead > 0)
            {
                int bytesRead = this.port.Read(incoming, 0, 256);
                var read = incoming.Take(bytesRead).ToArray();
                this.buffer.AddBytes(read);
            }
        }

        class IncomingBuffer
        {
            private int length;
            private readonly byte[] internalBuffer = new byte[MAX_LENGTH];

            public int Length
            {
                get
                {
                    return this.length;
                }
            }

            public void AddBytes(byte[] bytes)
            {
                if (bytes == null) throw new ArgumentNullException("bytes");
                if (bytes.Length + this.length > MAX_LENGTH) throw new ArgumentOutOfRangeException("bytes", "Total received too long.");
                Array.Copy(bytes, 0, this.internalBuffer, this.length, bytes.Length);
                this.length += bytes.Length;
            }

            public void Clear()
            {
                this.length = 0;
            }

            public byte[] GetBytes()
            {
                return this.internalBuffer.Take(this.length).ToArray();
            }
        }

        
        /// <summary>
        /// Wraps the given action inside of some exception handling, and 
        /// some checking to make sure the port is open before use.
        /// </summary>
        /// <param name="action">Action working with the serial port</param>
        private void serialPortTransaction(Action action)
        {
            try
            {
                makeSurePortIsOpen();
                action();
            }
            catch (Exception)
            {
                this.buffer.Clear();
                ClosePortIfOpen();
                throw;
            }
        }

        public void OpenPort()
        {
            ClosePortIfOpen();
            this.port = new SerialPort(this.comPortName, 19200, Parity.None, 8, StopBits.One);
            this.port.Open();
            this.port.DiscardOutBuffer();
            this.port.DiscardInBuffer();
        }

        private void makeSurePortIsOpen()
        {
            if (this.port == null || !this.port.IsOpen)
            {
                OpenPort();
            }
        }

        public void ClosePortIfOpen()
        {
            if (this.port != null && this.port.IsOpen)
            {
                try
                {
                    this.port.Close();
                }
                catch (Exception) { }
            }
            this.port = null;
        }

        public void Dispose()
        {
            ClosePortIfOpen();
        }
    }
}
