﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using PacketProcessor;
using System.IO;
using Common.Ini;
using System.Windows.Forms.DataVisualization.Charting;
using System.Threading;

namespace Viewer
{
    public partial class frmMain : Form
    {
        const double WATT_SECONDS_PER_KWH = 3600000.0;
        const int SERIES = TimeSlice.CHANNELS + 1;
        private static readonly string fullPathToDataFolder = getFullPathToDataFolder();
        private readonly Channel[] channels;

        public frmMain()
        {
            InitializeComponent();
            this.dtpStart.Value = DateTime.Today;
            this.dtpEnd.Value = this.dtpStart.Value.AddDays(1);
            this.channels = getChannels();
        }

        private static string getFullPathToDataFolder()
        {
            var defaultFolder = PacketProcessor.Program.FullPathToDataFolder();
            var iniFile = new IniFile();
            var dataFolder = iniFile.IniReadValue("FOLDERS", "OverrideViewerDataFolder");
            if (string.IsNullOrWhiteSpace(dataFolder))
            {
                return defaultFolder;
            }
            return dataFolder;
        }

        private void defineSeries()
        {
            // channels
            chart1.Series.Clear();
            for (var i = 0; i < SERIES; i++)
            {
                var series = chart1.Series.Add(this.channels[i].Name);
                series.ChartType = SeriesChartType.Line;
                series.XValueType = ChartValueType.String;
                series.YValueType = ChartValueType.Double;
                series.XValueMember = "Time";
                series.YValueMembers = string.Format("Value{0:00}", i);
            }
            this.dgvChannels.AutoGenerateColumns = false;
            this.dgvChannels.DataSource = channels;
        }

        private Channel[] getChannels()
        {
            var result = new Channel[TimeSlice.CHANNELS + 1];
            var iniFile = new IniFile();

            for (var i = 0; i < TimeSlice.CHANNELS; i++)
            {
                var key = string.Format("Channel{0:00}", i);
                var name = iniFile.IniReadValue("CHANNELS", key);
                result[i] = new Channel(this.onChannelSelectionChanged)
                    {
                        Index = i,
                        Name = name,
                    };
            }
            result[TimeSlice.CHANNELS] = new Channel(this.onChannelSelectionChanged)
            {
                Index = TimeSlice.CHANNELS,
                Name = "Other",
            };
            return result;
        }

        private void onChannelSelectionChanged()
        {
            for (var i = 0; i < SERIES; i++)
            {
                this.chart1.Series[i].Enabled = this.channels[i].Selected;
            }
            chart1.ChartAreas[0].RecalculateAxesScale();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            this.defineSeries();
            this.onChannelSelectionChanged();
            this.refreshChartData();
        }

        private void refreshChartData()
        {
            var start = this.dtpStart.Value;
            var end = this.dtpEnd.Value;
            var interval = end.Subtract(start);

            var originalCursor = this.Cursor;
            try
            {
                this.Cursor = Cursors.WaitCursor;
                int slicesPerPoint;
                string dateTimeFormat;
                bool kWh;
                if (interval.TotalMinutes <= 4 * 60)
                {
                    // 10 second line chart
                    slicesPerPoint = 1;
                    chart1.ChartAreas[0].AxisX.Interval = 360;
                    chart1.ChartAreas[0].AxisX.IntervalOffset = 1;
                    dateTimeFormat = "h tt";
                    this.setChartType(SeriesChartType.Line);
                    kWh = false;
                    chart1.ChartAreas[0].AxisY.Title = "Power (W)";
                }
                else if (interval.TotalMinutes <= 24 * 60)
                { 
                    // 1 minute line chart
                    slicesPerPoint = 6;
                    chart1.ChartAreas[0].AxisX.Interval = 60;
                    chart1.ChartAreas[0].AxisX.IntervalOffset = 1;
                    dateTimeFormat = "h tt";
                    this.setChartType(SeriesChartType.Line);
                    kWh = false;
                    chart1.ChartAreas[0].AxisY.Title = "Power (W)";
                }
                else if (interval.TotalMinutes <= 24 * 60 * 7)
                {
                    // hourly bar chart
                    slicesPerPoint = 6 * 60;
                    chart1.ChartAreas[0].AxisX.Interval = 3;
                    chart1.ChartAreas[0].AxisX.IntervalOffset = 0;
                    dateTimeFormat = "ddd h tt";
                    this.setChartType(SeriesChartType.StackedColumn);
                    kWh = true;
                    chart1.ChartAreas[0].AxisY.Title = "Energy (kWh)";
                }
                else 
                {
                    // daily bar chart
                    slicesPerPoint = 6 * 60 * 24;
                    chart1.ChartAreas[0].AxisX.Interval = 1;
                    chart1.ChartAreas[0].AxisX.IntervalOffset = 0;
                    dateTimeFormat = "ddd yyyy-MM-dd";
                    this.setChartType(SeriesChartType.StackedColumn);
                    kWh = true;
                    chart1.ChartAreas[0].AxisY.Title = "Energy (kWh)";
                }
                ulong[] onPeak_WS;
                ulong[] midPeak_WS;
                ulong[] offPeak_WS;
                var dataPoints = getDataPoints(start, end, slicesPerPoint, dateTimeFormat, kWh,
                    out onPeak_WS, out midPeak_WS, out offPeak_WS);
                chart1.DataSource = dataPoints;
                var onPeakOther = onPeak_WS[0] / WATT_SECONDS_PER_KWH;
                var midPeakOther = midPeak_WS[0] / WATT_SECONDS_PER_KWH;
                var offPeakOther = offPeak_WS[0] / WATT_SECONDS_PER_KWH;
                for (var ch = 0; ch < TimeSlice.CHANNELS; ch++)
                {
                    channels[ch].OnPeakEnergy_kWh = onPeak_WS[ch] / WATT_SECONDS_PER_KWH;
                    channels[ch].MidPeakEnergy_kWh = midPeak_WS[ch] / WATT_SECONDS_PER_KWH;
                    channels[ch].OffPeakEnergy_kWh = offPeak_WS[ch] / WATT_SECONDS_PER_KWH;
                    if (ch > 0)
                    {
                        onPeakOther -= channels[ch].OnPeakEnergy_kWh;
                        midPeakOther -= channels[ch].MidPeakEnergy_kWh;
                        offPeakOther -= channels[ch].OffPeakEnergy_kWh;
                    }
                }
                channels[TimeSlice.CHANNELS].OnPeakEnergy_kWh = onPeakOther;
                channels[TimeSlice.CHANNELS].MidPeakEnergy_kWh = midPeakOther;
                channels[TimeSlice.CHANNELS].OffPeakEnergy_kWh = offPeakOther;
                chart1.Invalidate();
                dgvChannels.Invalidate();
            }
            finally
            {
                this.Cursor = originalCursor;
            }
        }

        private void setChartType(SeriesChartType chartType)
        {
            for (var i = 0; i < SERIES; i++)
            {
                chart1.Series[i].ChartType = chartType;
            }
        }

        private static List<DataPoint> getDataPoints(
            DateTime start,
            DateTime end,
            int slicesPerPoint,
            string dateTimeFormat,
            bool kWh,
            out ulong[] onPeak_WS,
            out ulong[] midPeak_WS,
            out ulong[] offPeak_WS)
        {
            if (slicesPerPoint <= 0) throw new ArgumentOutOfRangeException("slicesPerPoint");

            onPeak_WS = new ulong[TimeSlice.CHANNELS];
            midPeak_WS = new ulong[TimeSlice.CHANNELS];
            offPeak_WS = new ulong[TimeSlice.CHANNELS];

            if (start >= end) return new List<DataPoint>();

            var result = new List<DataPoint>();
            var startDate = start.Date;
            var startOffset = start.Subtract(startDate);
            var startOffsetSlices = (long)(startOffset.TotalSeconds / TimeSlice.SECONDS_PER_SLICE);

            var endDate = end.Date;
            var endOffset = end.Subtract(endDate);
            var endOffsetSlices = (long)(endOffset.TotalSeconds / TimeSlice.SECONDS_PER_SLICE);

            for (var date = startDate; date <= endDate; date = date.AddDays(1))
            {
                var filename = SliceFileFinder.FilenameByDate(fullPathToDataFolder, date);
                var fileExists = File.Exists(filename);
                
                var fileInfo = new FileInfo(filename);
                var slicesInFile = fileExists
                    ? fileInfo.Length / TimeSlice.BYTES_PER_SLICE
                    : 0;
                var maxSlice = slicesInFile - 1;

                var firstSlice = date == startDate ? startOffsetSlices : 0;
                var lastSlice = date == endDate ? endOffsetSlices : maxSlice;
                if (lastSlice > maxSlice)
                {
                    lastSlice = maxSlice;
                }

                if (slicesInFile == 0 || firstSlice > maxSlice)
                {
                    var sliceCounter = 0;
                    var values = new double[SERIES];
                    for (var i = firstSlice; i <= lastSlice; i++)
                    {
                        sliceCounter++;
                        if (sliceCounter >= slicesPerPoint || i == lastSlice)
                        {
                            var pointEnding = date.AddSeconds(i * TimeSlice.SECONDS_PER_SLICE).ToString(dateTimeFormat);
                            result.Add(new DataPoint
                            {
                                Time = pointEnding,
                                Values = values,
                            });
                            sliceCounter = 0;
                        }
                    }
                    continue;
                }

                using (var fileStream = new FileStream(filename, FileMode.Open, FileAccess.Read))
                using (var reader = new BinaryReader(fileStream))
                {
                    var firstByte = (long)(firstSlice * TimeSlice.BYTES_PER_SLICE);
                    fileStream.Seek(firstByte, SeekOrigin.Begin);

                    var sliceCounter = 0;
                    var values = new double[SERIES];
                    for (var i = firstSlice; i <= lastSlice; i++)
                    {
                        var bytes = reader.ReadBytes(TimeSlice.BYTES_PER_SLICE);
                        var timeSlice = new TimeSlice(bytes);

                        var sliceTime = date.AddSeconds(i * TimeSlice.SECONDS_PER_SLICE);
                        var timeOfUse = sliceTime.ToTimeOfUse();
                        for (var ch = 0; ch < TimeSlice.CHANNELS; ch++)
                        {
                            switch (timeOfUse)
                            {
                                case TimeOfUse.OnPeak:
                                    onPeak_WS[ch] += timeSlice.WattSeconds(ch);
                                    break;
                                case TimeOfUse.MidPeak:
                                    midPeak_WS[ch] += timeSlice.WattSeconds(ch);
                                    break;
                                case TimeOfUse.OffPeak:
                                    offPeak_WS[ch] += timeSlice.WattSeconds(ch);
                                    break;
                            }
                        }

                        var divideBy = kWh ? WATT_SECONDS_PER_KWH : (double)TimeSlice.SECONDS_PER_SLICE * slicesPerPoint;

                        var mainPanel = timeSlice.WattSeconds(0) / divideBy;
                        var other = mainPanel;
                        for (var ch = 1; ch <= 13; ch++)
                        {
                            other -= timeSlice.WattSeconds(ch) / divideBy;
                            if (other < 0)
                            {
                                other = 0;
                            }
                        }
                        var thisSliceValues = new[] {
                                    timeSlice.WattSeconds(0) / divideBy,
                                    timeSlice.WattSeconds(1) / divideBy,
                                    timeSlice.WattSeconds(2) / divideBy,
                                    timeSlice.WattSeconds(3) / divideBy,
                                    timeSlice.WattSeconds(4) / divideBy,
                                    timeSlice.WattSeconds(5) / divideBy,
                                    timeSlice.WattSeconds(6) / divideBy,
                                    timeSlice.WattSeconds(7) / divideBy,
                                    timeSlice.WattSeconds(8) / divideBy,
                                    timeSlice.WattSeconds(9) / divideBy,
                                    timeSlice.WattSeconds(10) / divideBy,
                                    timeSlice.WattSeconds(11) / divideBy,
                                    timeSlice.WattSeconds(12) / divideBy,
                                    timeSlice.WattSeconds(13) / divideBy,
                                    other,
                                };
                        for (var s = 0; s < SERIES; s++)
                        {
                            values[s] += thisSliceValues[s];
                        }

                        sliceCounter++;
                        if (sliceCounter >= slicesPerPoint || i == lastSlice)
                        {
                            var pointTime = date.AddSeconds(i * TimeSlice.SECONDS_PER_SLICE);
                            var pointEnding = pointTime.ToString(dateTimeFormat);
                            result.Add(new DataPoint
                            {
                                Time = pointEnding,
                                Values = values,
                            });
                            values = new double[SERIES];
                            sliceCounter = 0;
                        }
                    }
                    reader.Close();
                    fileStream.Close();
                }
            }
            return result;
        }

        private void dgvChannels_CellMouseUp(object sender, DataGridViewCellMouseEventArgs e)
        {
            if (e.ColumnIndex == 0 && e.RowIndex > -1)
            {
                this.dgvChannels.EndEdit();
            }
        }

        private void cmdSelectBranches_Click(object sender, EventArgs e)
        {
            foreach (var channel in this.channels)
            {
                channel.SetSelected(true);
            }
            this.channels[0].SetSelected(false);
            this.onChannelSelectionChanged();
            this.dgvChannels.Refresh();
        }

        private void cmdSelectMainPanel_Click(object sender, EventArgs e)
        {
            foreach (var channel in this.channels)
            {
                channel.SetSelected(false);
            }
            this.channels[0].SetSelected(true);
            this.onChannelSelectionChanged();
            this.dgvChannels.Refresh();
        }

        private void cmdRefresh_Click(object sender, EventArgs e)
        {
            this.refreshChartData();
        }

        private void cmdToday_Click(object sender, EventArgs e)
        {
            this.dtpStart.Value = DateTime.Today;
            this.dtpEnd.Value = this.dtpStart.Value.AddDays(1);
            this.refreshChartData();
        }

        private void cmdYesterday_Click(object sender, EventArgs e)
        {
            this.dtpStart.Value = DateTime.Today.AddDays(-1);
            this.dtpEnd.Value = this.dtpStart.Value.AddDays(1);
            this.refreshChartData();
        }

        private void dgvChannels_CellFormatting(object sender, DataGridViewCellFormattingEventArgs e)
        {
            if (e.ColumnIndex >= 2)
            {
                double value = (double)(e.Value);
                e.Value = string.Format("{0:0.00}", value);
            }
        }
    }
}
